using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using Plux;

[Extension("Cerberus", OnReleased = "OnReleased", Singleton = false)]
[Plug("Discovery")]
[ParamValue("ExecuteInMainThread", true)]
[Slot("Analyzer", Multiple = true)]
public class Watcher : IDiscovery {
  private readonly List<DelayedFileSystemWatcher> assemblyWatcherList = new List<DelayedFileSystemWatcher>();

  private TypeBuilder typeBuilder;
  private ICollection<string> directories;

  public event DiscoveryEventHandler DiscoveryEvent;

  public void StartDiscovery(TypeBuilder builder) {
    StartDiscovery(builder, Runtime.BaseDirectories);
  }

  public void StartDiscovery(TypeBuilder builder, ICollection<string> watchDirectories) {
    typeBuilder = builder;
    directories = watchDirectories;
    if (Runtime.TaskQueue.IsBusy) {
      Runtime.TaskQueue.InvokeOnEmptied(Start);
    } else {
      Start();
    }
  }

  private void Start() {
    // TODOREMOVE: if runtime does not register event, it should crash here
    //if (DiscoveryEvent == null) {
    //  return;
    //}
    var plugins = new List<PluginInfo>();
    var contracts = new List<ContractInfo>();
    foreach (string dir in directories) {
      if (!Directory.Exists(dir)) {
        Runtime.GetExtensionInfo(this).LogEvent(MessageType.Error, "Directory not found: \"{0}\"", dir);
        continue;
      }
      string directory = Path.GetFullPath(dir);
      foreach (string fileName in Directory.GetFiles(directory)) {
        try {
          AddAssembly(fileName, contracts, plugins);
        } catch (Exception e) {
          Runtime.GetExtensionInfo(this).LogEvent(MessageType.Error, e.ToString());
        }
      }

      if (contracts.Count > 0 || plugins.Count > 0) {
        DiscoveryEvent(this, new DiscoveryEventArgs(DiscoveryChangeType.Add,
            plugins.ToArray(), contracts.ToArray()));
      }

      // TODO: *.dll not sufficient, should watch all supported types
      var watcher = new DelayedFileSystemWatcher(directory, "*.dll");
      watcher.NotifyFilter = NotifyFilters.FileName | NotifyFilters.LastWrite;
      watcher.Created += OnAssemblyDiscoveryEvent;
      watcher.Changed += OnAssemblyDiscoveryEvent;
      watcher.Deleted += OnAssemblyDiscoveryEvent;
      assemblyWatcherList.Add(watcher);

      foreach (DelayedFileSystemWatcher w in assemblyWatcherList) {
        w.EnableRaisingEvents = true;
      }
    }
  }

  public bool AnalyzeFile(string path, out List<ContractInfo> contractInfos, out List<PluginInfo> pluginInfos) {
    contractInfos = new List<ContractInfo>();
    pluginInfos = new List<PluginInfo>();
    var slot = Runtime.GetSlotInfo(this, "Analyzer");
    foreach (PlugInfo plug in slot.PluggedPlugInfos) {
      if (!DoesSupportFileType(plug.PlugTypeInfo, path)) {
        continue;
      }
      ContractInfo[] contracts;
      PluginInfo[] plugins;
      var analyzer = (IAnalyzer) plug.ExtensionInfo.Object;
      if (!analyzer.AnalyzeFile(path, typeBuilder, out contracts, out plugins)) {
        continue;
      }
      contractInfos.AddRange(contracts);
      pluginInfos.AddRange(plugins);
    }
    return contractInfos.Count > 0 || pluginInfos.Count > 0;
  }

  private static bool DoesSupportFileType(PlugTypeInfo plugType, string path) {
    var value = (string) plugType.GetParamValue("SupportedExtensions");
    if (value.Length == 0) {
      return true;
    }
    string fileExtension = Path.GetExtension(path);
    string[] types = value.Split(
        new[] {';'}, StringSplitOptions.RemoveEmptyEntries);
    foreach (string supports in types) {
      if (supports.Equals(fileExtension, StringComparison.OrdinalIgnoreCase)) {
        return true;
      }
    }
    return false;
  }

  private void OnAssemblyDiscoveryEvent(object source, FileSystemEventArgs e) {
    DiscoveryEventArgs args = null;
    switch (e.ChangeType) {
      case WatcherChangeTypes.Created:
      case WatcherChangeTypes.Changed:
      case WatcherChangeTypes.Renamed:
        var contract = new List<ContractInfo>();
        var plugins = new List<PluginInfo>();
        AddAssembly(e.FullPath, contract, plugins);
        if (contract.Count > 0 || plugins.Count > 0) {
          args = new DiscoveryEventArgs(DiscoveryChangeType.Add, plugins.ToArray(), contract.ToArray());
        }
        break;
      case WatcherChangeTypes.Deleted:
        PluginInfo[] existingPluginInfos = Runtime.GetPluginInfos(e.FullPath);
        ContractInfo[] existingContracts = Runtime.GetContractInfos(e.FullPath);
        if (existingPluginInfos.Length > 0 || existingContracts.Length > 0) {
          args = new DiscoveryEventArgs(DiscoveryChangeType.Remove, existingPluginInfos, existingContracts);
        }
        break;
    }
    if (args != null && DiscoveryEvent != null) {
      DiscoveryEvent(this, args);
    }
  }

  private void AddAssembly(string path, List<ContractInfo> contracts, List<PluginInfo> plugins) {
    PluginInfo[] pluginsInTypeStore = Runtime.GetPluginInfos(path);
    ContractInfo[] contractsInTypeStore = Runtime.GetContractInfos(path);
    var fileInfo = new FileInfo(path);
    if ((contractsInTypeStore.Length > 0 && contractsInTypeStore[0].RepositoryId.LastWriteTime == fileInfo.LastWriteTime) ||
        (pluginsInTypeStore.Length > 0 && pluginsInTypeStore[0].RepositoryId.LastWriteTime == fileInfo.LastWriteTime)) {
      return;
    }
    List<ContractInfo> newContracts;
    List<PluginInfo> newPlugins;
    AnalyzeFile(path, out newContracts, out newPlugins);
    contracts.AddRange(newContracts);
    plugins.AddRange(newPlugins);
  }

  public void OnReleased(object sender, ExtensionEventArgs args) {
    for (int i = assemblyWatcherList.Count - 1; i >= 0; i--) {
      assemblyWatcherList[i].EnableRaisingEvents = false;
      assemblyWatcherList[i].Dispose();
      assemblyWatcherList.RemoveAt(i);
    }
  }
}